@@ -16,12 +16,19 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
  * USA
  */
 
+#ifdef HAVE_CONFIG_H
+#  include <config.h>
+#endif
+
 #include <pyglib.h>
 #include "pygobject-private.h"
+#include "pyginterface.h"
+#include "pygparamspec.h"
+
 
 static void pygobject_dealloc(PyGObject *self);
 static int  pygobject_traverse(PyGObject *self, visitproc visit, void *arg);
@@ -31,6 +38,17 @@ static inline int pygobject_clear(PyGObject *self);
 static PyObject * pygobject_weak_ref_new(GObject *obj, PyObject *callback, PyObject *user_data);
 static inline PyGObjectData * pyg_object_peek_inst_data(GObject *obj);
 static PyObject * pygobject_weak_ref_new(GObject *obj, PyObject *callback, PyObject *user_data);
+static void pygobject_inherit_slots(PyTypeObject *type, PyObject *bases,
+				    gboolean check_for_present);
+static void pygobject_find_slot_for(PyTypeObject *type, PyObject *bases, int slot_offset,
+				    gboolean check_for_present);
+GType PY_TYPE_OBJECT = 0;
+GQuark pygobject_class_key;
+GQuark pygobject_class_init_key;
+GQuark pygobject_wrapper_key;
+GQuark pygobject_has_updated_constructor_key;
+GQuark pygobject_instance_data_key;
+
 
 /* -------------- class <-> wrapper manipulation --------------- */
 
@@ -83,7 +101,7 @@ pygobject_get_inst_data(PyGObject *self)
     {
         inst_data = pygobject_data_new();
 
-        inst_data->type = ((PyObject *)self)->ob_type;
+        inst_data->type = Py_TYPE(self);
         Py_INCREF((PyObject *) inst_data->type);
 
         g_object_set_qdata_full(self->obj, pygobject_instance_data_key,
@@ -166,6 +184,8 @@ typedef struct {
     guint index;
 } PyGPropsIter;
 
+PYGLIB_DEFINE_TYPE("gobject.GPropsIter", PyGPropsIter_Type, PyGPropsIter);
+
 static void
 pyg_props_iter_dealloc(PyGPropsIter *self)
 {
@@ -184,38 +204,6 @@ pygobject_props_iter_next(PyGPropsIter *iter)
     }
 }
 
-
-PyTypeObject PyGPropsIter_Type = {
-	PyObject_HEAD_INIT(NULL)
-	0,					/* ob_size */
-	"gobject.GPropsIter",			/* tp_name */
-	sizeof(PyGPropsIter),			/* tp_basicsize */
-	0,					/* tp_itemsize */
-	(destructor)pyg_props_iter_dealloc,	/* tp_dealloc */
-	0,					/* tp_print */
-	0,					/* tp_getattr */
-	0,					/* tp_setattr */
-	0,					/* tp_compare */
-	0,					/* tp_repr */
-	0,					/* tp_as_number */
-	0,					/* tp_as_sequence */
-	0,		       			/* tp_as_mapping */
-	0,					/* tp_hash */
-	0,					/* tp_call */
-	0,					/* tp_str */
-	0,					/* tp_getattro */
-	0,					/* tp_setattro */
-	0,					/* tp_as_buffer */
-	Py_TPFLAGS_DEFAULT,			/* tp_flags */
- 	"GObject properties iterator",		/* tp_doc */
-	0,					/* tp_traverse */
- 	0,					/* tp_clear */
-	0,					/* tp_richcompare */
-	0,					/* tp_weaklistoffset */
-	0,					/* tp_iter */
-	(iternextfunc)pygobject_props_iter_next, /* tp_iternext */
-};
-
 typedef struct {
     PyObject_HEAD
     /* a reference to the object containing the properties */
@@ -243,18 +231,19 @@ build_parameter_list(GObjectClass *class)
     GParamSpec **props;
     guint n_props = 0, i;
     PyObject *prop_str;
-    char *name;
     PyObject *props_list;
 
     props = g_object_class_list_properties(class, &n_props);
     props_list = PyList_New(n_props);
     for (i = 0; i < n_props; i++) {
+	char *name;
 	name = g_strdup(g_param_spec_get_name(props[i]));
 	/* hyphens cannot belong in identifiers */
 	g_strdelimit(name, "-", '_');
-	prop_str = PyString_FromString(name);
+	prop_str = _PyUnicode_FromString(name);
 	
 	PyList_SetItem(props_list, i, prop_str);
+	g_free(name);
     }
 
     if (props)
@@ -272,7 +261,7 @@ PyGProps_getattro(PyGProps *self, PyObject *attr)
     GValue value = { 0, };
     PyObject *ret;
 
-    attr_name = PyString_AsString(attr);
+    attr_name = _PyUnicode_AsString(attr);
     if (!attr_name) {
         PyErr_Clear();
         return PyObject_GenericGetAttr((PyObject *)self, attr);
@@ -347,6 +336,8 @@ set_property_from_pspec(GObject *obj,
     return TRUE;
 }
 
+PYGLIB_DEFINE_TYPE("gobject.GProps", PyGProps_Type, PyGProps);
+
 static int
 PyGProps_setattro(PyGProps *self, PyObject *attr, PyObject *pvalue)
 {
@@ -360,7 +351,7 @@ PyGProps_setattro(PyGProps *self, PyObject *attr, PyObject *pvalue)
 	return -1;
     }
 
-    attr_name = PyString_AsString(attr);
+    attr_name = _PyUnicode_AsString(attr);
     if (!attr_name) {
         PyErr_Clear();
         return PyObject_GenericSetAttr((PyObject *)self, attr, pvalue);
@@ -410,12 +401,14 @@ static Py_ssize_t
 PyGProps_length(PyGProps *self)
 {
     GObjectClass *class;
+    GParamSpec **props;
     guint n_props;
     
     class = g_type_class_ref(self->gtype);
-    g_object_class_list_properties(class, &n_props);
+    props = g_object_class_list_properties(class, &n_props);
     g_type_class_unref(class);
-    
+    g_free(props);
+
     return (Py_ssize_t)n_props;
 }
 
@@ -429,38 +422,7 @@ static PySequenceMethods _PyGProps_as_sequence = {
     0
 };
 
-PyTypeObject PyGProps_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,                                          /* ob_size */
-    "gobject.GProps",                           /* tp_name */
-    sizeof(PyGProps),                           /* tp_basicsize */
-    0,                                          /* tp_itemsize */
-    (destructor)PyGProps_dealloc,               /* tp_dealloc */
-    0,                                          /* tp_print */
-    0,                                          /* tp_getattr */
-    0,                                          /* tp_setattr */
-    0,                                          /* tp_compare */
-    0,                                          /* tp_repr */
-    0,                                          /* tp_as_number */
-    (PySequenceMethods*)&_PyGProps_as_sequence, /* tp_as_sequence */
-    0,                                          /* tp_as_mapping */
-    0,                                          /* tp_hash */
-    0,                                          /* tp_call */
-    0,                                          /* tp_str */
-    (getattrofunc)PyGProps_getattro,            /* tp_getattro */
-    (setattrofunc)PyGProps_setattro,            /* tp_setattro */
-    0,                                          /* tp_as_buffer */
-    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_GC,      /* tp_flags */
-    "The properties of the GObject accessible as "
-    "Python attributes.",                       /* tp_doc */
-    (traverseproc)pygobject_props_traverse,     /* tp_traverse */
-    0,                                          /* tp_clear */
-    0,                                          /* tp_richcompare */
-    0,                                          /* tp_weaklistoffset */
-    (getiterfunc)pygobject_props_get_iter,      /* tp_iter */
-    0,                                          /* tp_iternext */
-};
-
+PYGLIB_DEFINE_TYPE("gobject.GPropsDescr", PyGPropsDescr_Type, PyObject);
 
 static PyObject *
 pyg_props_descr_descr_get(PyObject *self, PyObject *obj, PyObject *type)
@@ -484,50 +446,6 @@ pyg_props_descr_descr_get(PyObject *self, PyObject *obj, PyObject *type)
     return (PyObject *) gprops;
 }
 
-PyTypeObject PyGPropsDescr_Type = {
-	PyObject_HEAD_INIT(NULL)
-	0,					/* ob_size */
-	"gobject.GPropsDescr",			/* tp_name */
-	0,					/* tp_basicsize */
-	0,					/* tp_itemsize */
-	0,	 				/* tp_dealloc */
-	0,					/* tp_print */
-	0,					/* tp_getattr */
-	0,					/* tp_setattr */
-	0,					/* tp_compare */
-	0,					/* tp_repr */
-	0,					/* tp_as_number */
-	0,					/* tp_as_sequence */
-	0,		       			/* tp_as_mapping */
-	0,					/* tp_hash */
-	0,					/* tp_call */
-	0,					/* tp_str */
-	0,					/* tp_getattro */
-	0,					/* tp_setattro */
-	0,					/* tp_as_buffer */
-	Py_TPFLAGS_DEFAULT,			/* tp_flags */
- 	0,					/* tp_doc */
-	0,					/* tp_traverse */
- 	0,					/* tp_clear */
-	0,					/* tp_richcompare */
-	0,					/* tp_weaklistoffset */
-	0,					/* tp_iter */
-	0,					/* tp_iternext */
-	0,					/* tp_methods */
-	0,					/* tp_members */
-	0,					/* tp_getset */
-	0,					/* tp_base */
-	0,					/* tp_dict */
-	pyg_props_descr_descr_get,		/* tp_descr_get */
-	0,					/* tp_descr_set */
-	0,					/* tp_dictoffset */
-	0,					/* tp_init */
-	0,					/* tp_alloc */
-	0,					/* tp_new */
-	0,               			/* tp_free */
-};
-
-
 /**
  * pygobject_register_class:
  * @dict: the module dictionary.  A reference to the type will be stored here.
@@ -586,13 +504,15 @@ pygobject_register_class(PyObject *dict, const gchar *type_name,
     } else
         bases = runtime_bases;
 
-    type->ob_type = PyGObject_MetaType;
+    Py_TYPE(type) = PyGObject_MetaType;
     type->tp_bases = bases;
     if (G_LIKELY(bases)) {
         type->tp_base = (PyTypeObject *)PyTuple_GetItem(bases, 0);
         Py_INCREF(type->tp_base);
     }
 
+    pygobject_inherit_slots(type, bases, TRUE);
+
     if (PyType_Ready(type) < 0) {
 	g_warning ("couldn't make the type `%s' ready", type->tp_name);
 	return;
@@ -603,7 +523,7 @@ pygobject_register_class(PyObject *dict, const gchar *type_name,
      */
     s = strrchr(type->tp_name, '.');
     if (s != NULL) {
-	mod_name = PyString_FromStringAndSize(type->tp_name, (int)(s - type->tp_name));
+	mod_name = _PyUnicode_FromStringAndSize(type->tp_name, (int)(s - type->tp_name));
 	PyDict_SetItemString(type->tp_dict, "__module__", mod_name);
 	Py_DECREF(mod_name);
     }
@@ -783,7 +703,7 @@ pygobject_new_with_interfaces(GType gtype)
 	type_name = g_strconcat(mod_name, ".", gtype_name, NULL);
     }
 
-    type = (PyTypeObject*)PyObject_CallFunction((PyObject *) py_parent_type->ob_type,
+    type = (PyTypeObject*)PyObject_CallFunction((PyObject *) Py_TYPE(py_parent_type),
                                                 "sNN", type_name, bases, dict);
     g_free(type_name);
 
@@ -810,6 +730,8 @@ pygobject_new_with_interfaces(GType gtype)
     type->tp_traverse = py_parent_type->tp_traverse;
     type->tp_clear = py_parent_type->tp_clear;
 
+    pygobject_inherit_slots(type, bases, FALSE);
+
     if (PyType_Ready(type) < 0) {
 	g_warning ("couldn't make the type `%s' ready", type->tp_name);
         pyglib_gil_state_release(state);
@@ -831,6 +753,97 @@ pygobject_new_with_interfaces(GType gtype)
     return type;
 }
 
+/* Pick appropriate value for given slot (at slot_offset inside
+ * PyTypeObject structure).  It must be a pointer, e.g. a pointer to a
+ * function.  We use the following heuristic:
+ *
+ * - Scan all types listed as bases of the type.
+ * - If for exactly one base type slot value is non-NULL and
+ *   different from that of 'object' and 'GObject', set current type
+ *   slot into that value.
+ * - Otherwise (if there is more than one such base type or none at
+ *   all) don't touch it and live with Python default.
+ *
+ * The intention here is to propagate slot from custom wrappers to
+ * wrappers created at runtime when appropriate.  We prefer to be on
+ * the safe side, so if there is potential collision (more than one
+ * custom slot value), we discard custom overrides altogether.
+ *
+ * When registering type with pygobject_register_class(), i.e. a type
+ * that has been manually created (likely with Codegen help),
+ * `check_for_present' should be set to TRUE.  In this case, the
+ * function will never overwrite any non-NULL slots already present in
+ * the type.  If `check_for_present' is FALSE, such non-NULL slots are
+ * though to be set by Python interpreter and so will be overwritten
+ * if heuristic above says so.
+ */
+static void
+pygobject_inherit_slots(PyTypeObject *type, PyObject *bases, gboolean check_for_present)
+{
+    static int slot_offsets[] = { offsetof(PyTypeObject, tp_richcompare),
+				  offsetof(PyTypeObject, tp_compare),
+				  offsetof(PyTypeObject, tp_hash),
+				  offsetof(PyTypeObject, tp_iter),
+				  offsetof(PyTypeObject, tp_repr),
+				  offsetof(PyTypeObject, tp_str),
+				  offsetof(PyTypeObject, tp_print) };
+    int i;
+
+    /* Happens when registering gobject.GObject itself, at least. */
+    if (!bases)
+	return;
+
+    for (i = 0; i < G_N_ELEMENTS(slot_offsets); ++i)
+	pygobject_find_slot_for(type, bases, slot_offsets[i], check_for_present);
+}
+
+static void
+pygobject_find_slot_for(PyTypeObject *type, PyObject *bases, int slot_offset,
+			gboolean check_for_present)
+{
+#define TYPE_SLOT(type)  (* (void **) (((char *) (type)) + slot_offset))
+
+    void *found_slot = NULL;
+    int num_bases = PyTuple_Size(bases);
+    int i;
+
+    if (check_for_present && TYPE_SLOT(type) != NULL) {
+	/* We are requested to check if there is any custom slot value
+	 * in this type already and there actually is.  Don't
+	 * overwrite it.
+	 */
+	return;
+    }
+
+    for (i = 0; i < num_bases; ++i) {
+	PyTypeObject *base_type = (PyTypeObject *) PyTuple_GetItem(bases, i);
+	void *slot = TYPE_SLOT(base_type);
+
+	if (slot == NULL)
+	    continue;
+	if (slot == TYPE_SLOT(&PyGObject_Type) ||
+	    slot == TYPE_SLOT(&PyBaseObject_Type))
+	    continue;
+
+	if (found_slot != NULL && found_slot != slot) {
+	    /* We have a conflict: more than one base use different
+	     * custom slots.  To be on the safe side, we bail out.
+	     */
+	    return;
+	}
+
+	found_slot = slot;
+    }
+
+    /* Only perform the final assignment if at least one base has a
+     * custom value.  Otherwise just leave this type's slot untouched.
+     */
+    if (found_slot != NULL)
+	TYPE_SLOT(type) = found_slot;
+
+#undef TYPE_SLOT
+}
+
 /**
  * pygobject_lookup_class:
  * @gtype: the GType of the GObject subclass.
@@ -973,6 +986,8 @@ pygobject_watch_closure(PyObject *self, GClosure *closure)
 
 /* -------------- PyGObject behaviour ----------------- */
 
+PYGLIB_DEFINE_TYPE("gobject.GObject", PyGObject_Type, PyGObject);
+
 static void
 pygobject_dealloc(PyGObject *self)
 {
@@ -984,7 +999,7 @@ pygobject_dealloc(PyGObject *self)
     pygobject_get_inst_data(self);
     pygobject_clear(self);
     /* the following causes problems with subclassed types */
-    /* self->ob_type->tp_free((PyObject *)self); */
+    /* Py_TYPE(self)->tp_free((PyObject *)self); */
     PyObject_GC_Del(self);
 }
 
@@ -1009,11 +1024,11 @@ pygobject_repr(PyGObject *self)
 
     g_snprintf(buf, sizeof(buf),
 	       "<%s object at 0x%lx (%s at 0x%lx)>",
-	       self->ob_type->tp_name,
+	       Py_TYPE(self)->tp_name,
 	       (long)self,
 	       self->obj ? G_OBJECT_TYPE_NAME(self->obj) : "uninitialized",
                (long)self->obj);
-    return PyString_FromString(buf);
+    return _PyUnicode_FromString(buf);
 }
 
 
@@ -1105,7 +1120,7 @@ pygobject_init(PyGObject *self, PyObject *args, PyObject *kwargs)
 	params = g_new0(GParameter, PyDict_Size(kwargs));
 	while (PyDict_Next (kwargs, &pos, &key, &value)) {
 	    GParamSpec *pspec;
-	    gchar *key_str = PyString_AsString(key);
+	    gchar *key_str = _PyUnicode_AsString(key);
 
 	    pspec = g_object_class_find_property (class, key_str);
 	    if (!pspec) {
@@ -1153,7 +1168,7 @@ pygobject__gobject_init__(PyGObject *self, PyObject *args, PyObject *kwargs)
     if (!G_IS_OBJECT(self->obj)) {                                           \
 	PyErr_Format(PyExc_TypeError,                                        \
                      "object at %p of type %s is not initialized",	     \
-                     self, self->ob_type->tp_name);                          \
+                     self, Py_TYPE(self)->tp_name);			     \
 	return NULL;                                                         \
     }
 
@@ -1211,13 +1226,13 @@ pygobject_get_properties(PyGObject *self, PyObject *args)
         GValue value = { 0 };
         PyObject *item;
 
-        if (!PyString_Check(py_property)) {
+        if (!_PyUnicode_Check(py_property)) {
             PyErr_SetString(PyExc_TypeError,
                             "Expected string argument for property.");
             return NULL;
         }
 
-        property_name = PyString_AsString(py_property);
+        property_name = _PyUnicode_AsString(py_property);
 
         pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(self->obj),
 					 property_name);
@@ -1281,6 +1296,7 @@ pygobject_set_properties(PyGObject *self, PyObject *args, PyObject *kwargs)
     Py_ssize_t      pos;
     PyObject        *value;
     PyObject        *key;
+    PyObject        *result = NULL;
 
     CHECK_GOBJECT(self);
 
@@ -1290,38 +1306,30 @@ pygobject_set_properties(PyGObject *self, PyObject *args, PyObject *kwargs)
     pos = 0;
 
     while (kwargs && PyDict_Next (kwargs, &pos, &key, &value)) {
-    gchar *key_str = PyString_AsString (key);
-    GParamSpec *pspec;
-    GValue gvalue ={ 0, };
+	gchar *key_str = _PyUnicode_AsString(key);
+	GParamSpec *pspec;
 
-    pspec = g_object_class_find_property (class, key_str);
-    if (!pspec) {
+	pspec = g_object_class_find_property(class, key_str);
+	if (!pspec) {
 	    gchar buf[512];
 
 	    g_snprintf(buf, sizeof(buf),
 		       "object `%s' doesn't support property `%s'",
 		       g_type_name(G_OBJECT_TYPE(self->obj)), key_str);
 	    PyErr_SetString(PyExc_TypeError, buf);
-	    return NULL;
+	    goto exit;
 	}
 
-	g_value_init(&gvalue, G_PARAM_SPEC_VALUE_TYPE(pspec));
-	if (pyg_value_from_pyobject(&gvalue, value)) {
-	    gchar buf[512];
-
-	    g_snprintf(buf, sizeof(buf),
-		       "could not convert value for property `%s'", key_str);
-	    PyErr_SetString(PyExc_TypeError, buf);
-	    return NULL;
-	}
-	g_object_set_property(G_OBJECT(self->obj), key_str, &gvalue);
-	g_value_unset(&gvalue);
+	if (!set_property_from_pspec(G_OBJECT(self->obj), key_str, pspec, value))
+	    goto exit;
     }
 
-    g_object_thaw_notify (G_OBJECT(self->obj));
+    result = Py_None;
 
-    Py_INCREF(Py_None);
-    return Py_None;
+ exit:
+    g_object_thaw_notify (G_OBJECT(self->obj));
+    Py_XINCREF(result);
+    return result;
 }
 
 static PyObject *
@@ -1435,7 +1443,7 @@ pygobject_connect(PyGObject *self, PyObject *args)
     if (!g_signal_parse_name(name, G_OBJECT_TYPE(self->obj),
 			     &sigid, &detail, TRUE)) {
 	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)self)),
 		     name);
 	return NULL;
     }
@@ -1484,7 +1492,7 @@ pygobject_connect_after(PyGObject *self, PyObject *args)
     if (!g_signal_parse_name(name, G_OBJECT_TYPE(self->obj),
 			     &sigid, &detail, TRUE)) {
 	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)self)),
 		     name);
 	return NULL;
     }
@@ -1533,7 +1541,7 @@ pygobject_connect_object(PyGObject *self, PyObject *args)
     if (!g_signal_parse_name(name, G_OBJECT_TYPE(self->obj),
 			     &sigid, &detail, TRUE)) {
 	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)self)),
 		     name);
 	return NULL;
     }
@@ -1582,7 +1590,7 @@ pygobject_connect_object_after(PyGObject *self, PyObject *args)
     if (!g_signal_parse_name(name, G_OBJECT_TYPE(self->obj),
 			     &sigid, &detail, TRUE)) {
 	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)self)),
 		     name);
 	return NULL;
     }
@@ -1681,7 +1689,7 @@ pygobject_emit(PyGObject *self, PyObject *args)
     if (!g_signal_parse_name(name, G_OBJECT_TYPE(self->obj),
 			     &signal_id, &detail, TRUE)) {
 	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)self)),
 		     name);
 	return NULL;
     }
@@ -1709,8 +1717,8 @@ pygobject_emit(PyGObject *self, PyObject *args)
 	if (pyg_value_from_pyobject(&params[i+1], item) < 0) {
 	    gchar buf[128];
 	    g_snprintf(buf, sizeof(buf),
-		"could not convert type %s to %s required for parameter %d",
-		item->ob_type->tp_name,
+		       "could not convert type %s to %s required for parameter %d",
+		       Py_TYPE(item)->tp_name,
 		g_type_name(G_VALUE_TYPE(&params[i+1])), i);
 	    PyErr_SetString(PyExc_TypeError, buf);
 
@@ -1757,7 +1765,7 @@ pygobject_stop_emission(PyGObject *self, PyObject *args)
     if (!g_signal_parse_name(signal, G_OBJECT_TYPE(self->obj),
 			     &signal_id, &detail, TRUE)) {
 	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)self)),
 		     signal);
 	return NULL;
     }
@@ -1821,9 +1829,9 @@ pygobject_chain_from_overridden(PyGObject *self, PyObject *args)
 	    gchar buf[128];
 
 	    g_snprintf(buf, sizeof(buf),
-		"could not convert type %s to %s required for parameter %d",
-		item->ob_type->tp_name,
-		g_type_name(G_VALUE_TYPE(&params[i+1])), i);
+		       "could not convert type %s to %s required for parameter %d",
+		       Py_TYPE(item)->tp_name,
+		       g_type_name(G_VALUE_TYPE(&params[i+1])), i);
 	    PyErr_SetString(PyExc_TypeError, buf);
 	    for (i = 0; i < query.n_params + 1; i++)
 		g_value_unset(&params[i]);
@@ -1905,7 +1913,7 @@ pygobject_disconnect_by_func(PyGObject *self, PyObject *args)
     closure = gclosure_from_pyfunc(self, pyfunc);
     if (!closure) {
 	PyErr_Format(PyExc_TypeError, "nothing connected to %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)pyfunc)));
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)pyfunc)));
 	return NULL;
     }
     
@@ -1914,7 +1922,7 @@ pygobject_disconnect_by_func(PyGObject *self, PyObject *args)
 						  0, 0,
 						  closure,
 						  NULL, NULL);
-    return PyInt_FromLong(retval);
+    return _PyLong_FromLong(retval);
 }
 
 static PyObject *
@@ -1937,7 +1945,7 @@ pygobject_handler_block_by_func(PyGObject *self, PyObject *args)
     closure = gclosure_from_pyfunc(self, pyfunc);
     if (!closure) {
 	PyErr_Format(PyExc_TypeError, "nothing connected to %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)pyfunc)));
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)pyfunc)));
 	return NULL;
     }
     
@@ -1946,7 +1954,7 @@ pygobject_handler_block_by_func(PyGObject *self, PyObject *args)
 					     0, 0,
 					     closure,
 					     NULL, NULL);
-    return PyInt_FromLong(retval);
+    return _PyLong_FromLong(retval);
 }
 
 static PyObject *
@@ -1969,7 +1977,7 @@ pygobject_handler_unblock_by_func(PyGObject *self, PyObject *args)
     closure = gclosure_from_pyfunc(self, pyfunc);
     if (!closure) {
 	PyErr_Format(PyExc_TypeError, "nothing connected to %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)pyfunc)));
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)pyfunc)));
 	return NULL;
     }
     
@@ -1978,7 +1986,7 @@ pygobject_handler_unblock_by_func(PyGObject *self, PyObject *args)
 					       0, 0,
 					       closure,
 					       NULL, NULL);
-    return PyInt_FromLong(retval);
+    return _PyLong_FromLong(retval);
 }
 
 static PyMethodDef pygobject_methods[] = {
@@ -2033,7 +2041,7 @@ pygobject_get_dict(PyGObject *self, void *closure)
 static PyObject *
 pygobject_get_refcount(PyGObject *self, void *closure)
 {
-    return PyInt_FromLong(self->obj->ref_count);
+    return _PyLong_FromLong(self->obj->ref_count);
 }
 
 static int
@@ -2057,57 +2065,9 @@ static PyGetSetDef pygobject_getsets[] = {
     { NULL, 0, 0 }
 };
 
-PyTypeObject PyGObject_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,					/* ob_size */
-    "gobject.GObject",			/* tp_name */
-    sizeof(PyGObject),			/* tp_basicsize */
-    0,					/* tp_itemsize */
-    /* methods */
-    (destructor)pygobject_dealloc,	/* tp_dealloc */
-    (printfunc)0,			/* tp_print */
-    (getattrfunc)0,			/* tp_getattr */
-    (setattrfunc)0,			/* tp_setattr */
-    (cmpfunc)pygobject_compare,		/* tp_compare */
-    (reprfunc)pygobject_repr,		/* tp_repr */
-    0,					/* tp_as_number */
-    0,					/* tp_as_sequence */
-    0,					/* tp_as_mapping */
-    (hashfunc)pygobject_hash,		/* tp_hash */
-    (ternaryfunc)0,			/* tp_call */
-    (reprfunc)0,			/* tp_str */
-    (getattrofunc)0,			/* tp_getattro */
-    (setattrofunc)pygobject_setattro,   /* tp_setattro */
-    0,					/* tp_as_buffer */
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE |
-    	Py_TPFLAGS_HAVE_GC,		/* tp_flags */
-    NULL, /* Documentation string */
-    (traverseproc)pygobject_traverse,	/* tp_traverse */
-    (inquiry)pygobject_clear,		/* tp_clear */
-    (richcmpfunc)0,			/* tp_richcompare */
-    offsetof(PyGObject, weakreflist),	/* tp_weaklistoffset */
-    (getiterfunc)0,			/* tp_iter */
-    (iternextfunc)0,			/* tp_iternext */
-    pygobject_methods,			/* tp_methods */
-    0,					/* tp_members */
-    pygobject_getsets,			/* tp_getset */
-    (PyTypeObject *)0,			/* tp_base */
-    (PyObject *)0,			/* tp_dict */
-    0,					/* tp_descr_get */
-    0,					/* tp_descr_set */
-    offsetof(PyGObject, inst_dict),	/* tp_dictoffset */
-    (initproc)pygobject_init,		/* tp_init */
-    (allocfunc)0,			/* tp_alloc */
-    (newfunc)0,				/* tp_new */
-    (freefunc)pygobject_free,		/* tp_free */
-    (inquiry)0,				/* tp_is_gc */
-    (PyObject *)0,			/* tp_bases */
-};
-
-
-  /* ------------------------------------ */
-  /* ****** GObject weak reference ****** */
-  /* ------------------------------------ */
+/* ------------------------------------ */
+/* ****** GObject weak reference ****** */
+/* ------------------------------------ */
 
 typedef struct {
     PyObject_HEAD
@@ -2117,6 +2077,8 @@ typedef struct {
     gboolean have_floating_ref;
 } PyGObjectWeakRef;
 
+PYGLIB_DEFINE_TYPE("gobject.GObjectWeakRef", PyGObjectWeakRef_Type, PyGObjectWeakRef);
+
 static int
 pygobject_weak_ref_traverse(PyGObjectWeakRef *self, visitproc visit, void *arg)
 {
@@ -2140,7 +2102,7 @@ pygobject_weak_ref_notify(PyGObjectWeakRef *self, GObject *dummy)
                 PyErr_Format(PyExc_TypeError,
                              "GObject weak notify callback returned a value"
                              " of type %s, should return None",
-                             retval->ob_type->tp_name);
+                             Py_TYPE(retval)->tp_name);
             Py_DECREF(retval);
             PyErr_Print();
         } else
@@ -2234,47 +2196,106 @@ pygobject_weak_ref_call(PyGObjectWeakRef *self, PyObject *args, PyObject *kw)
     }
 }
 
-PyTypeObject PyGObjectWeakRef_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,                                          /* ob_size */
-    "gobject.GObjectWeakRef",                   /* tp_name */
-    sizeof(PyGObjectWeakRef),                   /* tp_basicsize */
-    0,                                          /* tp_itemsize */
-    (destructor)pygobject_weak_ref_dealloc,     /* tp_dealloc */
-    0,                                          /* tp_print */
-    0,                                          /* tp_getattr */
-    0,                                          /* tp_setattr */
-    0,                                          /* tp_compare */
-    0,                                          /* tp_repr */
-    0,                                          /* tp_as_number */
-    0, 						/* tp_as_sequence */
-    0,                                          /* tp_as_mapping */
-    0,                                          /* tp_hash */
-    (ternaryfunc)pygobject_weak_ref_call,  	/*tp_call*/
-    0,                                          /* tp_str */
-    (getattrofunc)0,            		/* tp_getattro */
-    (setattrofunc)0,            		/* tp_setattro */
-    0,                                          /* tp_as_buffer */
-    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_GC,	/* tp_flags */
-    "A GObject weak reference",                 /* tp_doc */
-    (traverseproc)pygobject_weak_ref_traverse,  /* tp_traverse */
-    (inquiry)pygobject_weak_ref_clear,          /* tp_clear */
-    0,                                          /* tp_richcompare */
-    0,                                          /* tp_weaklistoffset */
-    0,      					/* tp_iter */
-    0,                                          /* tp_iternext */
-    pygobject_weak_ref_methods,			/* tp_methods */
-    0,						/* tp_members */
-    0,						/* tp_getset */
-    NULL,					/* tp_base */
-    0,						/* tp_dict */
-    0,						/* tp_descr_get */
-    0,						/* tp_descr_set */
-    0,						/* tp_dictoffset */
-    0,						/* tp_init */
-    0,						/* tp_alloc */
-    0,						/* tp_new */
-    0,						/* tp_free */
-    0,						/* tp_is_gc */
-    NULL,					/* tp_bases */
-};
+static gpointer
+pyobject_copy(gpointer boxed)
+{
+    PyObject *object = boxed;
+
+    Py_INCREF(object);
+    return object;
+}
+
+static void
+pyobject_free(gpointer boxed)
+{
+    PyObject *object = boxed;
+    PyGILState_STATE state;
+
+    state = pyglib_gil_state_ensure();
+    Py_DECREF(object);
+    pyglib_gil_state_release(state);
+}
+
+void
+pygobject_object_register_types(PyObject *d)
+{
+    PyObject *o, *descr;
+
+    pygobject_class_key = g_quark_from_static_string("PyGObject::class");
+    pygobject_class_init_key = g_quark_from_static_string("PyGObject::class-init");
+    pygobject_wrapper_key = g_quark_from_static_string("PyGObject::wrapper");
+    pygobject_has_updated_constructor_key =
+        g_quark_from_static_string("PyGObject::has-updated-constructor");
+    pygobject_instance_data_key = g_quark_from_static_string("PyGObject::instance-data");
+
+    /* GObject */
+    if (!PY_TYPE_OBJECT)
+	PY_TYPE_OBJECT = g_boxed_type_register_static("PyObject",
+						      pyobject_copy,
+						      pyobject_free);
+    PyGObject_Type.tp_dealloc = (destructor)pygobject_dealloc;
+    PyGObject_Type.tp_compare = (cmpfunc)pygobject_compare;
+    PyGObject_Type.tp_repr = (reprfunc)pygobject_repr;
+    PyGObject_Type.tp_hash = (hashfunc)pygobject_hash;
+    PyGObject_Type.tp_setattro = (setattrofunc)pygobject_setattro;
+    PyGObject_Type.tp_flags = (Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE |
+			       Py_TPFLAGS_HAVE_GC);
+    PyGObject_Type.tp_traverse = (traverseproc)pygobject_traverse;
+    PyGObject_Type.tp_clear = (inquiry)pygobject_clear;
+    PyGObject_Type.tp_weaklistoffset = offsetof(PyGObject, weakreflist);
+    PyGObject_Type.tp_methods = pygobject_methods;
+    PyGObject_Type.tp_getset = pygobject_getsets;
+    PyGObject_Type.tp_dictoffset = offsetof(PyGObject, inst_dict);
+    PyGObject_Type.tp_init = (initproc)pygobject_init;
+    PyGObject_Type.tp_free = (freefunc)pygobject_free;
+    PyGObject_Type.tp_alloc = PyType_GenericAlloc;
+    PyGObject_Type.tp_new = PyType_GenericNew;
+    pygobject_register_class(d, "GObject", G_TYPE_OBJECT,
+			     &PyGObject_Type, NULL);
+    PyDict_SetItemString(PyGObject_Type.tp_dict, "__gdoc__",
+			 pyg_object_descr_doc_get());
+    pyg_set_object_has_new_constructor(G_TYPE_OBJECT);
+
+    /* GProps */
+    PyGProps_Type.tp_dealloc = (destructor)PyGProps_dealloc;
+    PyGProps_Type.tp_as_sequence = (PySequenceMethods*)&_PyGProps_as_sequence;
+    PyGProps_Type.tp_getattro = (getattrofunc)PyGProps_getattro;
+    PyGProps_Type.tp_setattro = (setattrofunc)PyGProps_setattro;
+    PyGProps_Type.tp_flags = Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_GC;
+    PyGProps_Type.tp_doc = "The properties of the GObject accessible as "
+	"Python attributes.";
+    PyGProps_Type.tp_traverse = (traverseproc)pygobject_props_traverse;
+    PyGProps_Type.tp_iter = (getiterfunc)pygobject_props_get_iter;
+    if (PyType_Ready(&PyGProps_Type) < 0)
+        return;
+
+    /* GPropsDescr */
+    PyGPropsDescr_Type.tp_flags = Py_TPFLAGS_DEFAULT;
+    PyGPropsDescr_Type.tp_descr_get = pyg_props_descr_descr_get;
+    if (PyType_Ready(&PyGPropsDescr_Type) < 0)
+        return;
+    descr = PyObject_New(PyObject, &PyGPropsDescr_Type);
+    PyDict_SetItemString(PyGObject_Type.tp_dict, "props", descr);
+    PyDict_SetItemString(PyGObject_Type.tp_dict, "__module__",
+                        o=_PyUnicode_FromString("gobject._gobject"));
+    Py_DECREF(o);
+
+    /* GPropsIter */
+    PyGPropsIter_Type.tp_dealloc = (destructor)pyg_props_iter_dealloc;
+    PyGPropsIter_Type.tp_flags = Py_TPFLAGS_DEFAULT;
+    PyGPropsIter_Type.tp_doc = "GObject properties iterator";
+    PyGPropsIter_Type.tp_iternext = (iternextfunc)pygobject_props_iter_next;
+    if (PyType_Ready(&PyGPropsIter_Type) < 0)
+        return;
+
+    PyGObjectWeakRef_Type.tp_dealloc = (destructor)pygobject_weak_ref_dealloc;
+    PyGObjectWeakRef_Type.tp_call = (ternaryfunc)pygobject_weak_ref_call;
+    PyGObjectWeakRef_Type.tp_flags = Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_GC;
+    PyGObjectWeakRef_Type.tp_doc = "A GObject weak reference";
+    PyGObjectWeakRef_Type.tp_traverse = (traverseproc)pygobject_weak_ref_traverse;
+    PyGObjectWeakRef_Type.tp_clear = (inquiry)pygobject_weak_ref_clear;
+    PyGObjectWeakRef_Type.tp_methods = pygobject_weak_ref_methods;
+    if (PyType_Ready(&PyGObjectWeakRef_Type) < 0)
+        return;
+    PyDict_SetItemString(d, "GObjectWeakRef", (PyObject *) &PyGObjectWeakRef_Type);
+}
